from scipy.linalg import svd
from sklearn.utils.extmath import randomized_svd
import numpy as np


def tsvd(A, tol=0):
    U, S, V = svd(A, full_matrices=False)
    i = len(S)
    tol = tol*S[0]
    i = i - np.sum(np.nonzero(S[:i] <= tol))
    
    U = U[:,:i]
    S = S[:i]
    V = V.T[:,:i]
    return U, S, V



def rand_svd(A, tol, n_oversamples, random_state):
    n_components = min(A.shape)
    U, S, V = randomized_svd(A, n_components=n_components, n_oversamples=n_oversamples, random_state=random_state)
    i = len(S)
    tol = tol*S[0]
    i = i - np.sum(np.nonzero(S[:i] <= tol))
    
    U = U[:,:i]
    S = S[:i]
    V = V.T[:,:i]
    return U, S, V


from dataclasses import dataclass
import numpy as np

@dataclass
class tree:
    leafnode: bool = False
    topnode: bool = False
    blockmLeft: int = -1
    blockmRight: int = -1
    blocknLeft: int = -1
    blocknRight: int = -1
    Bl : np.ndarray = None
    Bu : np.ndarray = None
    leftChild : object = None
    rightChild : object = None
    RLeft : np.ndarray = None
    RRight : np.ndarray = None
    WLeft : np.ndarray = None
    WRight : np.ndarray = None
    U : np.ndarray = None
    V : np.ndarray = None
    D : np.ndarray = None



import numpy as np
from dataclasses import dataclass
from .tree import tree


@dataclass
class treeG:
    leafnode : bool = False
    topnode : bool = False
    vbl : tree = None
    vbr : tree = None
    gl : tree = None
    gr : tree = None




def _down_sweep(trg, hsstree, f, x):
    if not hsstree.leafnode:
        if hsstree.topnode:
            fl = np.matmul(hsstree.Bu, trg.gr)
            fr = np.matmul(hsstree.Bl, trg.gl)
        else:
            fl = np.matmul(hsstree.Bu, trg.gr) + np.matmul(hsstree.RLeft, f)
            fr = np.matmul(hsstree.Bl, trg.gl) + np.matmul(hsstree.RRight, f)

        bl = _down_sweep(trg.vbl, hsstree.leftChild, fl, x[:hsstree.blocknLeft,:])
        br = _down_sweep(trg.vbr, hsstree.rightChild, fr, x[hsstree.blocknLeft:,:])

        b = np.vstack((bl, br))

    else:
        b = np.matmul(hsstree.D, x) + np.matmul(hsstree.U, f)
    return b




def _up_sweep(hsstree, x):
    trg = treeG()
    g = None

    if not hsstree.leafnode: 
        trg.vbl, trg.gl = _up_sweep(hsstree.leftChild, x[:hsstree.blocknLeft,:])
        trg.vbr, trg.gr = _up_sweep(hsstree.rightChild, x[hsstree.blocknLeft:,:])
        # trg.vbl, trg.gl, trg.vbr, trg.gr = vbl, gl, vbr, gr

        if not hsstree.topnode:
            g = np.matmul(hsstree.WLeft.T, trg.gl) + np.matmul(hsstree.WRight.T, trg.gr)    

    else:  
        g = np.matmul(hsstree.V.T, x)

    return trg, g





def hss_matmul(hsstree, x):
    if not hsstree.leafnode:           
        vb, _ = _up_sweep(hsstree, x)
        
        f = None
        b = _down_sweep(vb, hsstree, f, x)
    else:
        b = np.matmul(hsstree.D, x)
    return b



import numpy as np
from .factor import *
from .tree import tree


def _allExept(block, nmax):
    return np.concatenate((np.arange(block[0], dtype=int), 
                           np.arange(block[-1]+1, nmax, dtype=int)), axis=None)    


def _hss_build_iter(A, tol, maxdepth, p=1, depth=0, blockm=None, blockn=None, U=None, V=None, lk_approx='svd', n_oversamples=10, random_state=0):
    depth = depth + 1
    seed = np.random.uniform(0, 1, 1)
    parentNode = len(blockm)>2 and len(blockn)>2 and p>seed and depth<maxdepth

    if parentNode:
        hsstree = tree(leafnode=False, topnode=False,
                       blockmLeft=int(np.floor((blockm[-1] - blockm[0]+1)/2)),
                       blockmRight=int(np.ceil((blockm[-1] - blockm[0]+1)/2)),
                       blocknLeft=int(np.floor((blockn[-1] - blockn[0]+1)/2)),
                       blocknRight=int(np.ceil((blockn[-1] - blockn[0]+1)/2))
                    )
        
        blockmLeft = np.arange(blockm[0], blockm[0]+hsstree.blockmLeft)
        blockmRight = np.arange(blockm[0]+hsstree.blockmLeft, blockm[-1]+1)
        blocknLeft = np.arange(blockn[0], blockn[0]+hsstree.blocknLeft)
        blocknRight = np.arange(blockn[0]+hsstree.blocknLeft, blockn[-1]+1)
        
        removeblockmLeft = _allExept(blockmLeft, A.shape[0])
        removeblockmRight = _allExept(blockmRight, A.shape[0])
        removeblocknLeft = _allExept(blocknLeft, A.shape[1])
        removeblocknRight = _allExept(blocknRight, A.shape[1])

        if lk_approx == 'randsvd':
            ULeft, _, _ = rand_svd(A[blockmLeft][:, removeblocknLeft], tol, n_oversamples, random_state)
            URight, _, _  = rand_svd(A[blockmRight][:, removeblocknRight], tol, n_oversamples, random_state)
            U12, S3, VLeft = rand_svd(A[removeblockmLeft][:, blocknLeft], tol, n_oversamples, random_state)
            U21, S4, VRight = rand_svd(A[removeblockmRight][:, blocknRight], tol, n_oversamples, random_state)

        else:
            ULeft, _, _ = tsvd(A[blockmLeft][:, removeblocknLeft], tol)
            URight, _, _  = tsvd(A[blockmRight][:, removeblocknRight], tol)
            U12, S3, VLeft = tsvd(A[removeblockmLeft][:, blocknLeft], tol)
            U21, S4, VRight = tsvd(A[removeblockmRight][:, blocknRight], tol)

        U12Ind = np.arange(blockm[0], blockm[0] + URight.shape[0])
        U21Ind = np.arange(blockm[0], blockm[0] + ULeft.shape[0])
        Bl = np.matmul(np.matmul(URight.T, U12[U12Ind]), np.diag(S3))
        Bu = np.matmul(np.matmul(ULeft.T, U21[U21Ind]), np.diag(S4))
        
        hsstree.Bl=Bl
        hsstree.Bu=Bu

        hsstree.leftChild = _hss_build_iter(A, tol, maxdepth, p, depth, blockmLeft, blocknLeft, ULeft, VLeft, n_oversamples, random_state)
        hsstree.rightChild = _hss_build_iter(A, tol, maxdepth, p, depth, blockmRight, blocknRight, URight, VRight, n_oversamples, random_state)

        if depth != 1:
            hsstree.RLeft = np.linalg.solve(ULeft, U[:hsstree.blockmLeft,:])
            hsstree.RRight = np.linalg.solve(URight, U[hsstree.blockmLeft:,:])

            hsstree.WLeft = np.linalg.solve(VLeft, V[:hsstree.blocknLeft,:])
            hsstree.WRight = np.linalg.solve(VRight, V[hsstree.blocknLeft:,:])
        else:
            hsstree.topnode = True

    else:
        hsstree = tree(leafnode=True, topnode=False)
        if depth != 1:
            hsstree.U = U
            hsstree.V = V
        else:
            hsstree.topnode=True
        
        hsstree.D = A[blockm][:, blockn]

    return hsstree


class hss(object):

    def __init__(self, tol=0, maxdepth=np.inf, p=1, lk_approx='svd', n_oversamples=10, random_state=0):
        self.tol = tol
        self.maxdepth = maxdepth
        self.p = p
        self.lk_approx = lk_approx
        self.n_oversamples = n_oversamples
        self.random_state = random_state

    def build(self, A):
        m, n = A.shape
        blockm = np.arange(m)
        blockn = np.arange(n)

        return _hss_build_iter(A, 
                               tol=self.tol,
                               maxdepth=self.maxdepth, 
                               p=self.p, 
                               blockm=blockm, 
                               blockn=blockn,
                               lk_approx=self.lk_approx, 
                               n_oversamples=self.n_oversamples,
                               random_state=self.random_state
                               )


    

 


import setuptools
import numpy
import platform
import importlib
import logging

PACKAGE_NAME = "pyhss"
VERSION = "0.0.1"
SETREQUIRES=["numpy"]
MAINTAINER="No name"
EMAIL="xinyechenai@gmail.com"
INREUIRES=["numpy>=1.7.2"]


AUTHORS="Xinye Chen"

with open("README.md", 'r') as f:
    long_description = f.read()

ext_errors = (ModuleNotFoundError, IOError, SystemExit)
logging.basicConfig()
log = logging.getLogger(__file__)

if platform.python_implementation() == "PyPy":
    NUMPY_MIN_VERSION = "1.19.2"
else:
    NUMPY_MIN_VERSION = "1.17.2"
   
    
metadata = {"name":PACKAGE_NAME,
            "packages":[PACKAGE_NAME],
            "version":VERSION,
            "setup_requires":SETREQUIRES,
            "install_requires":INREUIRES,
            "include_dirs":[numpy.get_include()],
            "long_description":long_description,
            "author":AUTHORS,
            "maintainer":MAINTAINER,
            "author_email":EMAIL,
            "classifiers":[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: Unix",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            ],
            "maintainer_email":EMAIL,
            "description":"A python package for hierarchically semiseparable matrix representation",
            "long_description_content_type":'text/markdown',
            "url":"https://github.com/chenxinye/pyhss.git",
            "license":'MIT License'
}
            

class InvalidVersion(ValueError):
    """raise invalid version error"""

    
def check_package_status(package, min_version):
    """
    check whether given package.
    """
    package_status = {}
    try:
        module = importlib.import_module(package)
        package_version = module.__version__
        package_status["up_to_date"] = package_version >= min_version
        package_status["version"] = package_version
    except ImportError:
        traceback.print_exc()
        package_status["up_to_date"] = False
        package_status["version"] = ""

    req_str = "pychop requires {} >= {}.\n".format(package, min_version)

    if package_status["up_to_date"] is False:
        if package_status["version"]:
            raise ImportError(
                "Your installation of {} {} is out-of-date.\n{}".format(
                    package, package_status["version"], req_str
                )
            )
        else:
            raise ImportError(
                "{} is not installed.\n{}{}".format(package, req_str)
            )


def setup_package():
    check_package_status("numpy", NUMPY_MIN_VERSION)
    
    setuptools.setup(
        **metadata
    )
    


if __name__ == "__main__":
    try:
        setup_package()
    except ext_errors as ext:
        log.warn(ext)
        log.warn("failure Installation.")
